////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.facebook;

import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import org.json.JSONObject;


/***********************************************************************************************
Details about a single Facebook friend
***********************************************************************************************/
public class FacebookFriend implements Comparable<FacebookFriend> {

  // RETRIEVED FROM https://graph.facebook.com/<user_id>
  /** The Facebook user ID **/
  long userID = -1;
  /** The users first name **/
  String firstName = null;
  /** The users middle name **/
  String middleName = null;
  /** The users last name **/
  String lastName = null;
  /** The gender of this user **/
  String gender = null;
  /** The users username **/
  String username = null;
  /** The users bio message **/
  String bio = null;
  /** The users birth date **/
  String birthday = null;
  /** The email address of this user **/
  String emailAddress = null;
  /** The home city of this user **/
  String homeCity = null;
  /** The current location of this user **/
  String location = null;
  /** The users political stance **/
  String politics = null;
  /** The users relationship status (single, married, etc) **/
  String relationshipStatus = null;
  /** The users religious stance **/
  String religion = null;
  /** The users significant other **/
  String marriedTo = null;
  /** The users website **/
  String website = null;

  // RETRIEVED FROM https://graph.facebook.com/<user_id>/picture
  /** The URL to the photo of this user **/
  URL photo = null;

  // RETRIEVED FROM https://graph.facebook.com/<user_id>/statuses
  /** The last status message from the user **/
  String statusMessage = null;
  /** The time of the last status message from the user **/
  long statusMessageTime = 0;

  // RETRIEVED FROM https://graph.facebook.com/<user_id>/checkins
  /** The last location this user checked-in to **/
  String checkin = null;
  /** The last time this user checked-in to a venue **/
  long checkinTime = 0;

  
  
  /***********************************************************************************************
  Creates a new Facebook friend
  ***********************************************************************************************/
  public FacebookFriend(){

  }

  /***********************************************************************************************
  Creates a new Facebook friend from JSON data
  @param friend the JSON data for the friend
  ***********************************************************************************************/
  public FacebookFriend(JSONObject friend){
    toComponent(friend);
  }


  /***********************************************************************************************
  Creates a new Facebook friend
  @param userID the Facebook User ID
  ***********************************************************************************************/
  public FacebookFriend(long userID){
    this.userID = userID;
  }


  public String getBio(){
    return bio;
  }


  public String getBirthday(){
    return birthday;
  }


  public String getCheckin(){
    return checkin;
  }


  public long getCheckinTime(){
    return checkinTime;
  }


  public String getEmailAddress(){
    return emailAddress;
  }
  
  
  public boolean hasEmailAddress(){
    return ((emailAddress != null) && !emailAddress.equals(""));
  }
  
  
  public boolean hasFirstName(){
    return ((firstName != null) && !firstName.equals(""));
  }
  
  
  public boolean hasLastName(){
    return ((lastName != null) && !lastName.equals(""));
  }


  public String getFirstName(){
    return firstName;
  }


  public String getGender(){
    return gender;
  }


  public String getHomeCity(){
    return homeCity;
  }


  public String getLastName(){
    return lastName;
  }


  public String getLocation(){
    return location;
  }


  public String getMarriedTo(){
    return marriedTo;
  }


  public String getMiddleName(){
    return middleName;
  }


  public URL getPhoto(){
    return photo;
  }


  public String getPolitics(){
    return politics;
  }


  public String getRelationshipStatus(){
    return relationshipStatus;
  }


  public String getReligion(){
    return religion;
  }


  public String getStatusMessage(){
    return statusMessage;
  }


  public long getStatusMessageTime(){
    return statusMessageTime;
  }


  /***********************************************************************************************
  Gets the user ID of this user
  @return the <i>userID</i>
  ***********************************************************************************************/
  public long getUserID(){
    return userID;
  }


  public String getUsername(){
    return username;
  }


  public String getWebsite(){
    return website;
  }


  public void setBio(String bio){
    this.bio = bio;
  }


  public void setBirthday(String birthday){
    if (birthday == null || birthday.equals("")){
      this.birthday = "";
      return; // v1.3 bug fix - added missing return to prevent dumps
    }
    
    // Convert the birthday format from Facebook to Android format
    int firstSlashPosition = birthday.indexOf('/');
    int secondSlashPosition = birthday.indexOf('/',firstSlashPosition+1);
    
    if (firstSlashPosition > 0){
      if (secondSlashPosition > firstSlashPosition+1){
        // MM/DD/YYYY --> YYYY-MM-DD
        String month = birthday.substring(0,firstSlashPosition);
        String day = birthday.substring(firstSlashPosition+1,secondSlashPosition);
        String year = birthday.substring(secondSlashPosition+1);
        this.birthday = year + "-" + month + "-" + day;
      }
      else {
        // MM/DD --> 0000-MM-DD
        String month = birthday.substring(0,firstSlashPosition);
        String day = birthday.substring(firstSlashPosition+1);
        //this.birthday = "--" + month + "-" + day;
        //this.birthday = month + "-" + day;
        this.birthday = "0000-" + month + "-" + day;
      }
    }
    else {
      this.birthday = "";
    }
  }


  public void setCheckin(String checkin){
    this.checkin = checkin;
  }


  public void setCheckinTime(long checkinTime){
    this.checkinTime = checkinTime;
  }


  /***********************************************************************************************
  Sets the last check-in time for this user. Converts the String into a Long
  @param checkinTime the last check-in time for this user
  ***********************************************************************************************/
  public void setCheckinTime(String checkinTime){

    try {
      SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
      Date date = dateFormatter.parse(checkinTime);
      Calendar calendar = Calendar.getInstance();
      calendar.setTime(date);
      this.checkinTime = calendar.getTimeInMillis();
    }
    catch (Throwable t) {
      this.checkinTime = 0;
    }

  }


  public void setEmailAddress(String emailAddress){
    this.emailAddress = emailAddress;
  }


  public void setFirstName(String firstName){
    this.firstName = firstName;
  }


  public void setGender(String gender){
    this.gender = gender;
  }


  public void setHomeCity(String homeCity){
    this.homeCity = homeCity;
  }


  public void setLastName(String lastName){
    this.lastName = lastName;
  }


  public void setLocation(String location){
    this.location = location;
  }


  public void setMarriedTo(String marriedTo){
    this.marriedTo = marriedTo;
  }


  public void setMiddleName(String middleName){
    this.middleName = middleName;
  }


  public void setPhoto(URL photo){
    this.photo = photo;
  }


  public void setPolitics(String politics){
    this.politics = politics;
  }


  public void setRelationshipStatus(String relationshipStatus){
    this.relationshipStatus = relationshipStatus;
  }


  public void setReligion(String religion){
    this.religion = religion;
  }


  public void setStatusMessage(String statusMessage){
    this.statusMessage = statusMessage;
  }


  public void setStatusMessageTime(long statusMessageTime){
    this.statusMessageTime = statusMessageTime;
  }


  /***********************************************************************************************
  Sets the last status message time for this user. Converts the String into a Long
  @param statusMessageTime the last status message time for this user
  ***********************************************************************************************/
  public void setStatusMessageTime(String statusMessageTime){

    try {
      SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
      Date date = dateFormatter.parse(statusMessageTime);
      Calendar calendar = Calendar.getInstance();
      calendar.setTime(date);
      this.statusMessageTime = calendar.getTimeInMillis();
    }
    catch (Throwable t) {
      this.statusMessageTime = 0;
    }

  }


  /***********************************************************************************************
  Sets the User ID of this user
  @param userID the User ID of this user
  ***********************************************************************************************/
  public void setUserID(long userID){
    this.userID = userID;
  }


  public void setUsername(String username){
    this.username = username;
  }


  public void setWebsite(String website){
    this.website = website;
  }


  /***********************************************************************************************
  Populates the FacebookFriend from JSON data
  @param friend the JSON data for the friend
  ***********************************************************************************************/
  public void toComponent(JSONObject friend){
    try {
      this.userID = friend.optLong("id",-1);
    }
    catch (Throwable t) {}

  }



  /***********************************************************************************************
  Compares this friend to <i>another</i> based on their userID
  @param another the friend to compare to
  @return a positive or negative number representing the order
  ***********************************************************************************************/
  public int compareTo(FacebookFriend another){
    long otherID = another.getUserID();
    if (otherID == userID){
      return 0;
    }
    else if (otherID < userID){
      return 1;
    }
    else {
      return -1;
    }
  }

}
