////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.facebook;

import java.net.URI;
import java.util.Arrays;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.watto.android.Settings;
import org.watto.program.android.sync.facebook.authenticator.FacebookAuthenticator;
import org.watto.program.android.sync.facebook.http.HttpMultipartMode;
import org.watto.program.android.sync.facebook.http.MultipartEntity;
import org.watto.program.android.sync.facebook.http.content.StringBody;
import android.content.ContentResolver;
import android.content.Context;
import android.database.Cursor;
import android.provider.ContactsContract.CommonDataKinds.Email;
import android.provider.ContactsContract.CommonDataKinds.StructuredName;
import android.provider.ContactsContract.Data;


/***********************************************************************************************
 Provides utility methods for communicating with the backend server
***********************************************************************************************/
public class FacebookQuery {
  
  
  /** The Context for performing user queries against the android database **/
  static Context context = null;
  
  public static void setContext(Context appContext){
    context = appContext;
  }


  /***********************************************************************************************
  Fetches the list of friend data updates from the server
  @return list The list of updates received from the server.
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] getFriends(){
    try {

      if (!FacebookAuthenticator.credentialsKnown()) {
        // The authentication is invalid, so return nothing
        return new FacebookFriend[0];
      }
      
      
      // Determine whether to get a small or large profile photo
      String photoSize = Settings.get("PhotoSize");
      if (photoSize == null || photoSize.equals("") || photoSize.equals("small")){
        photoSize = "";
      }
      else if (photoSize.equals("medium")){
        photoSize = "?type=normal";
      }
      else if (photoSize.equals("large")){
        photoSize = "?type=large";
      }
      else {
        photoSize = "";
      }
      

      /*
      String friendsURL = "https://graph.facebook.com/";
      // Sign the request with the access token
      HttpPost request = new HttpPost(friendsURL);

      String batchData = "[{ \"method\": \"GET\", \n" + 
                            "\"name\" : \"get-friends\", \n" + 
                            "\"relative_url\": \"me/friends\", \n" +
                            "}, \n" +
                          "{ \"method\": \"GET\", \n" + 
                            "\"relative_url\": \"?ids={result=get-friends:$.data.*.id}\" \n" +
                          "}]";

      // Create a new MultipartEntity for holding the POST data
      MultipartEntity entity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE);
      StringBody batchDataBody = new StringBody(batchData);
      entity.addPart("batch",batchDataBody);
      entity.addPart("access_token",new StringBody(FacebookAuthenticator.authClient.getAccessToken()));

      request.setEntity(entity);

      // Send the request
      DefaultHttpClient client = new DefaultHttpClient();
      HttpResponse response = client.execute(request);

      // Get the response code
      if (response.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
        // Success
        String responseMessage = EntityUtils.toString(response.getEntity());
        
        String batchObject = new JSONArray(responseMessage).getJSONObject(1).optString("body","");
        JSONObject friendsObject = new JSONObject(batchObject);
        */
        
      String batchRequest = "[{ \"method\": \"GET\", \n" + 
                               "\"name\" : \"get-friends\", \n" + 
                               "\"relative_url\": \"me/friends\", \n" +
                               "}, \n" +
                             "{ \"method\": \"GET\", \n" + 
                               "\"relative_url\": \"?ids={result=get-friends:$.data.*.id}\" \n" +
                             "}]";
      
        JSONArray responses = queryServerBatch(batchRequest);
        JSONObject responseObject = new JSONObject(responses.getJSONObject(1).optString("body",""));


        JSONArray friendsArray = responseObject.names();

        int friendCount = friendsArray.length();
        FacebookFriend[] friends = new FacebookFriend[friendCount];
        long[] ids = new long[friendCount];
        
        // quick-exit if there was an error or no friends were found, so we don't go and
        // request checkins or statuses later on
        if (friendCount <= 0){
          return new FacebookFriend[0];
        }

        for (int i = 0;i < friendCount;i++) {
          FacebookFriend friend = new FacebookFriend();

          JSONObject detailsObject = responseObject.getJSONObject(friendsArray.optString(i));

          // User ID
          long userId = detailsObject.optLong("id",-1);
          friend.setUserID(userId);
          ids[i] = userId;

          // Name
          friend.setFirstName(detailsObject.optString("first_name",""));
          friend.setMiddleName(detailsObject.optString("middle_name",""));
          friend.setLastName(detailsObject.optString("last_name",""));
          friend.setUsername(detailsObject.optString("username",""));

          // Gender
          friend.setGender(detailsObject.optString("gender",""));

          // Bio
          friend.setBio(detailsObject.optString("bio",""));

          // Birthday
          friend.setBirthday(detailsObject.optString("birthday",""));

          // Email and Website
          friend.setEmailAddress(detailsObject.optString("email",""));
          friend.setWebsite(detailsObject.optString("website",""));

          // Hometown and Location
          try {
            friend.setHomeCity(detailsObject.getJSONObject("hometown").optString("name",""));
          }
          catch (JSONException e) {}
          try {
            friend.setLocation(detailsObject.getJSONObject("location").optString("name",""));
          }
          catch (JSONException e) {}

          // Politics and Religion
          friend.setPolitics(detailsObject.optString("political",""));
          friend.setReligion(detailsObject.optString("religion",""));

          // Married Status
          friend.setRelationshipStatus(detailsObject.optString("relationship_status",""));
          try {
            friend.setMarriedTo(detailsObject.getJSONObject("significant_other").optString("name",""));
          }
          catch (JSONException e) {}

          // get the picture for this user
          friend.setPhoto(URI.create("https://graph.facebook.com/" + userId + "/picture" + photoSize).toURL());

          friends[i] = friend;

        }
        
        
        // sort the friends so that the getCheckins and getStatuses will return the answers
        // in the right order (because the Facebook server re-orders them in numerical order).
        java.util.Arrays.sort(friends);
        
        //getCheckins(friends,ids);
        //getStatuses(friends,ids);
        getCheckins(friends);
        getStatuses(friends);

        return friends;

      /*
      }
      */

    }
    catch (Throwable t) {
      // Error in the response - maybe the server is down or the response message has changed
      String error = "error";
    }
    return new FacebookFriend[0];
  }
  
  

  
  
  /***********************************************************************************************
  Gets the names for all contacts on the user's phone
  @return The list of names
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] getNamesForContacts(){
    try {
      
      ContentResolver resolver = context.getContentResolver();
      String[] columns = new String[]{Data.DATA2,Data.DATA3};
      String whereClause = Data.MIMETYPE + "='" + StructuredName.CONTENT_ITEM_TYPE + "'";
      Cursor c = resolver.query(Data.CONTENT_URI,columns,whereClause,null,null);
      
      int nameCount = c.getCount();
      FacebookFriend[] friends = new FacebookFriend[nameCount];

      try {
        int currentFriend = 0;
        while (c.moveToNext()) {
          FacebookFriend friend = new FacebookFriend(); 
          
          // filters out people missing their first or last name
          String firstName = c.getString(0);
          if (firstName == null || firstName.equals("")){
            continue;
          }
          friend.setFirstName(firstName);
          
          String lastName = c.getString(1);
          if (lastName == null || lastName.equals("")){
            continue;
          }
          friend.setLastName(lastName);
          
          // create the friend
          friends[currentFriend] = friend;
          currentFriend++;
          }
        
          // Shrink the array if some were left blank due to a missing first/last name
          if (currentFriend < nameCount){
            FacebookFriend[] oldFriends = friends;
            friends = new FacebookFriend[currentFriend];
            System.arraycopy(oldFriends,0,friends,0,currentFriend);
          }
        }
      finally {
        c.close();
        }
      
      return friends;
 
    }
    catch (Throwable t){
      String error = "error";
    }
    return new FacebookFriend[0];
  }
  
  
  /***********************************************************************************************
  Gets the Facebook user IDs for each name, if they are a friend and are on facebook
  @return The Facebook user IDs for the names
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] getFacebookUserIdsForNames(FacebookFriend[] friends){
    try {
      
      if (friends == null || friends.length <= 0){
        return friends;
      }
            
      JSONObject response = queryServer("https://graph.facebook.com/me/friends");
      if (response == null){
        return friends;
      }
      
      JSONArray responseArray = response.getJSONArray("data");
      int responseCount = responseArray.length();
      
      
      // Calculate hashes for all the friends[] so we can find them quickly
      int friendCount = friends.length;
      int[] friendHashes = new int[friendCount];
      for (int i=0;i<friendCount;i++){
        FacebookFriend friend = friends[i];
        friendHashes[i] = (friend.getFirstName() + " " + friend.getLastName()).hashCode();
      }
      
            
      // Go through each FacebookFriend and see whether this friend was returned in the response
      for (int i = 0;i < responseCount;i++) {
        JSONObject singleResponse = responseArray.getJSONObject(i);
        
        String name = singleResponse.optString("name",null);
        // find the friend that has the same first and last name
        int nameHash = name.hashCode();
        for (int j=0;j<friendCount;j++){
          if (nameHash == friendHashes[j]){
            // found a match - verify by checking the actual first and last name
            FacebookFriend friend = friends[j];
            if (name.equals(friend.getFirstName() + " " + friend.getLastName())){
              // set the user ID for this friend
              friend.setUserID(singleResponse.optLong("id",-1));
              break;
            }
          }
        }
      }


      return friends;
      
    }
    catch (Throwable t){
      String error = "error";
    }
    return new FacebookFriend[0];
  }
  
  
  
  /***********************************************************************************************
  Gets the email addresses for all contacts on the user's phone
  @return The list of email addresses
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] getEmailAddressesForContacts(){
    try {
      
      ContentResolver resolver = context.getContentResolver();
      String[] columns = new String[]{Data.DATA1};
      String whereClause = Data.MIMETYPE + "='" + Email.CONTENT_ITEM_TYPE + "'";
      Cursor c = resolver.query(Data.CONTENT_URI,columns,whereClause,null,null);
      
      int emailCount = c.getCount();
      FacebookFriend[] friends = new FacebookFriend[emailCount];

      try {
        int currentFriend = 0;
        while (c.moveToNext()) {
          FacebookFriend friend = new FacebookFriend(); 
          friend.setEmailAddress(c.getString(0));
          friends[currentFriend] = friend;
          currentFriend++;
          }
        }
      finally {
        c.close();
        }
      
      return friends;
 
    }
    catch (Throwable t){
      String error = "error";
    }
    return new FacebookFriend[0];
  }
  
  

  public static int BATCH_TYPE_SEARCH_EMAIL = 1;
  public static int BATCH_TYPE_GET_CHECKINS = 2;
  public static int BATCH_TYPE_GET_STATUSES = 3;
  
  /***********************************************************************************************
  Creates the batch requests of the given <i>type</i> for querying the server
  ***********************************************************************************************/
  public static String[] createBatchRequests(int type, FacebookFriend[] friends){
    try {
      
      int friendCount = friends.length;
            
      // max 20 requests per batch
      int blockCount = friendCount / 20;
      int extraCount = friendCount ^ 20;
      if (extraCount > 0) {
        blockCount++;
      }

      String[] batchRequests = new String[blockCount];
      int currentFriend = 0;
      for (int i = 0;i < blockCount;i++) {
        String batchData = "[";

        for (int j = 0;j < 20;j++) {
          if (currentFriend >= friendCount) {
            break;
          }

          if (j > 0) {
            batchData += ",";
          }

          if (type == BATCH_TYPE_SEARCH_EMAIL){
            batchData += "{ \"method\": \"GET\", \n" + 
                           "\"relative_url\": \"search?q=" + friends[currentFriend].getEmailAddress() + "&type=user\", \n" + 
                          "}\n";
          }
          else if (type == BATCH_TYPE_GET_CHECKINS){
            batchData += "{ \"method\": \"GET\", \n" + 
                           "\"relative_url\": \"" + friends[currentFriend].getUserID() + "/checkins?limit=1&fields=place,created_time\", \n" + 
                          "}\n";
          }
          else if (type == BATCH_TYPE_GET_STATUSES){
            batchData += "{ \"method\": \"GET\", \n" + 
                           "\"relative_url\": \"" + friends[currentFriend].getUserID() + "/statuses?limit=1&fields=message,updated_time\", \n" + 
                          "}\n";
          }
          
          currentFriend++;
        }
        batchData += "]";

        batchRequests[i] = batchData;
      }
      
      return batchRequests;
    }
    catch (Throwable t){
    }
    return new String[0];
  }
  
  
  /***********************************************************************************************
  Sends a batch request to the server and returns the responses as a <code>JSONArray</code>
  @param batchRequest the request
  @return the responses
  ***********************************************************************************************/
  public static JSONArray queryServerBatch(String batchRequest){
    try {
      String batchURL = "https://graph.facebook.com/";
      HttpPost request = new HttpPost(batchURL);

      // Create a new MultipartEntity for holding the POST data
      MultipartEntity entity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE);
      StringBody batchDataBody = new StringBody(batchRequest);
      entity.addPart("batch",batchDataBody);
      entity.addPart("access_token",new StringBody(FacebookAuthenticator.authClient.getAccessToken()));

      request.setEntity(entity);

      // Send the request
      DefaultHttpClient client = new DefaultHttpClient();
      HttpResponse response = client.execute(request);

      // Get the response code
      if (response.getStatusLine().getStatusCode() != HttpStatus.SC_OK) {
        // Failure
        return new JSONArray();
      }

      // Success
      String responseMessage = EntityUtils.toString(response.getEntity());
      JSONArray responses = new JSONArray(responseMessage);
      return responses;
    }
    catch (Throwable t){
    }
    return new JSONArray();
  }
  
  
  /***********************************************************************************************
  Checks the userIds of the <i>friends</i> to ensure that they are all friends of the user. If not,
  the friend is removed, as it is likely to be a hijacked email address.
  @return The validated Facebook friends
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] validateFriendUserIds(FacebookFriend[] friends){
    try {
      
      if (friends == null || friends.length <= 0){
        return friends;
      }
      
      
        JSONObject response = queryServer("https://graph.facebook.com/me/friends?fields=id");
        if (response == null){
          // Failure somewhere, so skip this step
          return friends;
        }
        
        JSONArray friendList = response.optJSONArray("data");
        if (friendList == null){
          return friends;
        }
        
        // Get an array of friend userIds
        int friendListCount = friendList.length();
        long[] friendIds = new long[friendListCount];
        
        for (int i=0;i<friendListCount;i++){
          try {
            friendIds[i] = friendList.getJSONObject(i).getLong("id");
          }
          catch (Throwable t){
          }
        }
        
        // sort for binary lookup
        Arrays.sort(friendIds);
        
        // now check each friend to see if it has a friendIds id
        int friendCount = friends.length;
        
        FacebookFriend[] newFriends = new FacebookFriend[friendCount];
        int newFriendCount = 0;
        
        
        for (int i=0;i<friendCount;i++){
          FacebookFriend friend = friends[i];
          long userId = friend.getUserID();
          
          boolean found = (Arrays.binarySearch(friendIds,userId) >= 0);
          
          if (found){
            // copy into the new array
            newFriends[newFriendCount] = friend;
            newFriendCount++;
          }
          // else ignore it - not a valid friend
          
        }
        
        
        if (newFriendCount == friendCount){
          // all friends were valid
          return friends;
        }
        else {
          // shrink the array to the new size
          friends = new FacebookFriend[newFriendCount];
          System.arraycopy(newFriends,0,friends,0,newFriendCount);
          return friends;
        }
      
    }
    catch (Throwable t){
      String error = "error";
    }
    return friends;
  }
  
  
  /***********************************************************************************************
  Gets the Facebook user IDs for each email address, if they are a friend and are on facebook
  @return The Facebook user IDs for the email addresses
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] getFacebookUserIdsForEmailAddresses(FacebookFriend[] friends){
    try {
      
      if (friends == null || friends.length <= 0){
        return friends;
      }
      
      String[] batchRequests = createBatchRequests(BATCH_TYPE_SEARCH_EMAIL,friends);
      int batchCount = batchRequests.length;

      // call the batch URLs and populate the friends
      int friendCount = friends.length;
      int currentFriend = 0;
      for (int c = 0;c < batchCount;c++) {
        JSONArray responses = queryServerBatch(batchRequests[c]);
        if (responses == null || responses.length() <= 0){
          // Failure somewhere, so skip and move on to the next batch
          continue;
        }
        
        for (int b = 0;b < 20;b++) { // 20 to a batch
          if (currentFriend >= friendCount) {
            break;
          }

          try {
            JSONObject responseObject = new JSONObject(responses.getJSONObject(b).optString("body",""));
            JSONArray responseDataArray = responseObject.getJSONArray("data");
            
            if (responseDataArray.length() >= 1) {
              // User ID
              JSONObject foundUser = responseDataArray.getJSONObject(0);
              friends[currentFriend].setUserID(foundUser.optLong("id",-1));
            }
          }
          catch (JSONException e) {
            // Ignore any errors trying to get the IDs
          }

          currentFriend++;
        }
      }
      
      return friends;
      
    }
    catch (Throwable t){
      String error = "error";
    }
    return new FacebookFriend[0];
  }
  
  
  /***********************************************************************************************
  Removes any <i>friends</i> that do not have a userID attached to them (because they didn't
  match a facebook friend). Also joins duplicate friends together (if they have the same userID)
  @param oldFriends the friends to filter
  @return the filtered friends 
  ***********************************************************************************************/
  public static FacebookFriend[] filterEmptyUserIds(FacebookFriend[] oldFriends){
    try {
      
      if (oldFriends == null){
        return oldFriends;
      }
      
      java.util.Arrays.sort(oldFriends);
      
      int friendCount = oldFriends.length;
      if (friendCount <= 1){
        return oldFriends;
      }
      
      FacebookFriend[] friends = new FacebookFriend[friendCount];
      int currentFriend = 0;
      
      FacebookFriend previousFriend = null;
      for (int i=0;i<friendCount;i++){
        FacebookFriend oldFriend = oldFriends[i];
        long userID = oldFriend.getUserID();
        
        if (userID <= -1){
          // No user ID, so not a match to a Facebook friend
          continue;
        }
        
        if (previousFriend != null){
          // see if the previous friend userID is the same as this userID
          if (previousFriend.getUserID() == userID){
            // Match, so the same user, so join them together
            if (oldFriend.hasEmailAddress()){
              previousFriend.setEmailAddress(oldFriend.getEmailAddress());
            }
            if (oldFriend.hasFirstName()){
              previousFriend.setFirstName(oldFriend.getFirstName());
            }
            if (oldFriend.hasLastName()){
              previousFriend.setLastName(oldFriend.getLastName());
            }
          }
          else {
            // different, so add this new friend to the array
            friends[currentFriend] = oldFriend;
            currentFriend++;
            
            previousFriend = oldFriend;
          }
        }
        else {
          // no previous friend, so just add it straight to the new array
          friends[currentFriend] = oldFriend;
          currentFriend++;
          
          previousFriend = oldFriend;
        }
        
      }
      
    // Shrink the array to the right size
    FacebookFriend[] newFriends = new FacebookFriend[currentFriend];
    System.arraycopy(friends,0,newFriends,0,currentFriend);
    
    return newFriends;
      
    }
    catch (Throwable t){
    }
    return oldFriends;
  }
  
  
  /***********************************************************************************************
  Removes any <i>friends</i> that do not have a first name attached to them.
  If you have an email address for a contact, however someone else registered that email address
  in Facebook but it is NOT your friend, sometimes Facebook will still return that userId in the
  search() even though that person is not your friend. When we go through to populate the detailed
  friend data, we will not get detailed information about this person, because they are not your
  friend, so if they don't have something fundamental such as First Name then we will filter them
  out as this situation has probably occurred.
  @param oldFriends the friends to filter
  @return the filtered friends 
  ***********************************************************************************************/
  public static FacebookFriend[] filterEmptyFirstNames(FacebookFriend[] oldFriends){
    try {
      
      int friendCount = oldFriends.length;
      int currentFriend = 0;
      
      FacebookFriend[] friends = new FacebookFriend[friendCount];
      
      for (int i=0;i<friendCount;i++){
        FacebookFriend oldFriend = oldFriends[i];
        String firstName = oldFriend.getFirstName();
        
        if (firstName == null || firstName.equals("")){
          // No first name, so remove the user (skip them)
          continue;
        }
        else {
          // valid, so add to the new array
          friends[currentFriend] = oldFriend;
          currentFriend++;
        }
      }
      
    if (currentFriend == friendCount){
      // the new array is fully populated (no filtered-out entries), so just return it
      return friends;
    }
      
    // Shrink the array to the right size
    FacebookFriend[] newFriends = new FacebookFriend[currentFriend];
    System.arraycopy(friends,0,newFriends,0,currentFriend);
    
    return newFriends;
      
    }
    catch (Throwable t){
    }
    return oldFriends;
  }
  
  
  /***********************************************************************************************
  Merges multiple <code>FacebookFriend</code> arrays together
  @return the merged array
  ***********************************************************************************************/
  public static FacebookFriend[] mergeFriendArrays(FacebookFriend[] emailFriends,FacebookFriend[] nameFriends){
    try {
      int emailCount = emailFriends.length;
      int nameCount = nameFriends.length;
      
      FacebookFriend[] friends = new FacebookFriend[emailCount+nameCount];
      System.arraycopy(emailFriends,0,friends,0,emailCount);
      System.arraycopy(nameFriends,0,friends,emailCount,nameCount);
      
      return friends;
    }
    catch (Throwable t){
      return emailFriends;
    }
  }
  
  
  /***********************************************************************************************
  Fetches the list of friend data updates from the server, but only for the friends that are
  currently stored in the users' contact address book
  @return list The list of updates received from the server.
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static FacebookFriend[] getCurrentFriendsOnly(){
    try {
      
      if (!FacebookAuthenticator.credentialsKnown()) {
        // The authentication is invalid, so return nothing
        return new FacebookFriend[0];
      }
      
      // Gets the friends based on their email addresses
      FacebookFriend[] emailFriends = getEmailAddressesForContacts();
      emailFriends = getFacebookUserIdsForEmailAddresses(emailFriends);
      
      // Gets the friends based on their first and last name
      FacebookFriend[] nameFriends = getNamesForContacts();
      nameFriends = getFacebookUserIdsForNames(nameFriends);
      
      // Merge the email and name arrays together
      FacebookFriend[] friends = mergeFriendArrays(emailFriends,nameFriends);
      // Sorts and filters the list of friends to only those with userIds
      friends = filterEmptyUserIds(friends);
      
      // Check that these userIDs are actually my friends!
      // This filters out people that have hijacked email addresses but actually
      // a friend of the user.
      friends = validateFriendUserIds(friends);
      
      if (friends == null || friends.length <= 0){
        // No valid friends found that match email addresses or names, so get everyone instead
        return getFriends();
      }
      
      
      // Determine whether to get a small or large profile photo
      String photoSize = Settings.get("PhotoSize");
      if (photoSize == null || photoSize.equals("") || photoSize.equals("small")){
        photoSize = "";
      }
      else if (photoSize.equals("medium")){
        photoSize = "?type=normal";
      }
      else if (photoSize.equals("large")){
        photoSize = "?type=large";
      }
      else {
        photoSize = "";
      }
      
      
      
      int friendCount = friends.length;
      
      
      // quick-exit if there was an error or no friends were found, so we don't go and
      // request checkins or statuses later on
      if (friendCount <= 0){
        return new FacebookFriend[0];
      }
      
      
      long[] ids = new long[friendCount]; // for use with getCheckins() and getStatuses()
      String userIdString = ""; // for batch-querying the server
      for (int i=0;i<friendCount;i++){
        if (i > 0){
          userIdString += ",";
        }
        long userId = friends[i].getUserID();
        ids[i] = userId;
        userIdString += userId;
      }

      String batchRequest = "[{ \"method\": \"GET\", \n" + 
                               "\"relative_url\": \"?ids=" + userIdString + "\" \n" +
                             "}]";

      JSONArray responses = queryServerBatch(batchRequest);
        
      // NOTE: THIS IS DIFFERENT TO getFriends() (the index is 0 instead of 1)!!!
      JSONObject responseObject = new JSONObject(responses.getJSONObject(0).optString("body",""));
      
      //JSONArray friendsArray = responseObject.names();
        

        for (int i = 0;i < friendCount;i++) {
          FacebookFriend friend = friends[i];

          JSONObject detailsObject = responseObject.getJSONObject(""+friend.getUserID());

          // User ID (already known!)
          //long userId = detailsObject.optLong("id",-1);
          //friend.setUserID(userId);

          // Name
          friend.setFirstName(detailsObject.optString("first_name",""));
          friend.setMiddleName(detailsObject.optString("middle_name",""));
          friend.setLastName(detailsObject.optString("last_name",""));
          friend.setUsername(detailsObject.optString("username",""));

          // Gender
          friend.setGender(detailsObject.optString("gender",""));

          // Bio
          friend.setBio(detailsObject.optString("bio",""));

          // Birthday
          friend.setBirthday(detailsObject.optString("birthday",""));

          // Email and Website
          //friend.setEmailAddress(detailsObject.optString("email",""));
          friend.setWebsite(detailsObject.optString("website",""));

          // Hometown and Location
          try {
            friend.setHomeCity(detailsObject.getJSONObject("hometown").optString("name",""));
          }
          catch (JSONException e) {}
          try {
            friend.setLocation(detailsObject.getJSONObject("location").optString("name",""));
          }
          catch (JSONException e) {}

          // Politics and Religion
          friend.setPolitics(detailsObject.optString("political",""));
          friend.setReligion(detailsObject.optString("religion",""));

          // Married Status
          friend.setRelationshipStatus(detailsObject.optString("relationship_status",""));
          try {
            friend.setMarriedTo(detailsObject.getJSONObject("significant_other").optString("name",""));
          }
          catch (JSONException e) {}

          // get the picture for this user
          friend.setPhoto(URI.create("https://graph.facebook.com/" + friend.getUserID() + "/picture" + photoSize).toURL());
        }
        
        
        // Filter out anyone who does not have a first name.
        friends = filterEmptyFirstNames(friends);
        
                
        //getCheckins(friends,ids);
        //getStatuses(friends,ids);
        getCheckins(friends);
        getStatuses(friends);

        return friends;

    }
    catch (Throwable t) {
      // Error in the response - maybe the server is down or the response message has changed
      String error = "error";
    }
    return new FacebookFriend[0];
  }
  
  
  
  
  
  /***********************************************************************************************
  Fetches the checkins for each friend
  @param friends the friends
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static void getCheckins(FacebookFriend[] friends){
    try {

      if (friends == null || friends.length == 0) {
        return;
      }
  
      String[] batchRequests = createBatchRequests(BATCH_TYPE_GET_CHECKINS,friends);
      int batchCount = batchRequests.length;

      // call the batch URLs and populate the friends
      int friendCount = friends.length;
      int currentFriend = 0;
      for (int c = 0;c < batchCount;c++) {
        JSONArray responses = queryServerBatch(batchRequests[c]);
        if (responses == null || responses.length() <= 0) {
          // Failure somewhere, so skip and move on to the next batch
          continue;
        }

        for (int b = 0;b < 20;b++) { // 20 to a batch
          if (currentFriend >= friendCount) {
            break;
          }

          try {
            JSONObject responseObject = new JSONObject(responses.getJSONObject(b).optString("body",""));
            JSONArray responseDataArray = responseObject.getJSONArray("data");

            if (responseDataArray.length() >= 1) {
              FacebookFriend friend = friends[currentFriend];

              JSONObject checkinWrapper = responseDataArray.getJSONObject(0);
              // Location
              try {
                friend.setCheckin(checkinWrapper.getJSONObject("place").optString("name",""));
              }
              catch (JSONException e) {}

              // Timestamp (2011-07-02T00:45:38+0000)
              friend.setCheckinTime(checkinWrapper.optString("created_time",""));

            }
          }
          catch (JSONException e) {
            // Ignore any errors trying to get the IDs
          }

          currentFriend++;
        }
      }
    }
    catch (Throwable t) {
      // Error in the response - maybe the server is down or the response message has changed
      String error = "error";
    }
  }
  
  
  /***********************************************************************************************
  Fetches the current status message for each friend
  @param friends the friends
  ***********************************************************************************************/
  @SuppressWarnings("unused")
  public static void getStatuses(FacebookFriend[] friends){
    try {

      if (friends == null || friends.length == 0) {
        return;
      }
  
      String[] batchRequests = createBatchRequests(BATCH_TYPE_GET_STATUSES,friends);
      int batchCount = batchRequests.length;

      // call the batch URLs and populate the friends
      int friendCount = friends.length;
      int currentFriend = 0;
      for (int c = 0;c < batchCount;c++) {
        JSONArray responses = queryServerBatch(batchRequests[c]);
        if (responses == null || responses.length() <= 0) {
          // Failure somewhere, so skip and move on to the next batch
          continue;
        }

        for (int b = 0;b < 20;b++) { // 20 to a batch
          if (currentFriend >= friendCount) {
            break;
          }

          try {
            JSONObject responseObject = new JSONObject(responses.getJSONObject(b).optString("body",""));
            JSONArray responseDataArray = responseObject.getJSONArray("data");

            if (responseDataArray.length() >= 1) {
              FacebookFriend friend = friends[currentFriend];

              JSONObject statusWrapper = responseDataArray.getJSONObject(0);

              // Message
              friend.setStatusMessage(statusWrapper.optString("message",""));

              // Timestamp (2011-07-02T00:45:38+0000)
              friend.setStatusMessageTime(statusWrapper.optString("updated_time",""));
            }
          }
          catch (JSONException e) {
            // Ignore any errors trying to get the IDs
          }

          currentFriend++;
        }
      }
    }
    catch (Throwable t) {
      // Error in the response - maybe the server is down or the response message has changed
      String error = "error";
    }
  }

  /***********************************************************************************************
  Fetches the checkins for each friend
  @param friends the friends
  @param ids the user ids for each friend
  ***********************************************************************************************/
  /*public static void oldGetCheckins(FacebookFriend[] friends,long[] ids){
    try {

      if (friends == null || friends.length == 0 || ids == null || ids.length == 0) {
        return;
      }


      int friendCount = friends.length;
      if (friendCount != ids.length) {
        return;
      }

      // Create the batch requests (max of 20 per batch request)
      int blockCount = friendCount / 20;
      int extraCount = friendCount ^ 20;
      if (extraCount > 0) {
        blockCount++;
      }

      String[] batchRequests = new String[blockCount];
      int currentFriend = 0;
      for (int i = 0;i < blockCount;i++) {
        String batchData = "[";

        for (int j = 0;j < 20;j++) {
          if (currentFriend >= friendCount) {
            break;
          }

          if (j > 0) {
            batchData += ",";
          }

          batchData += "{ \"method\": \"GET\", \n" + 
                         "\"relative_url\": \"" + ids[currentFriend] + "/checkins?limit=1&fields=place,created_time\", \n" + 
                        "}\n";
          currentFriend++;
        }
        batchData += "]";

        batchRequests[i] = batchData;
      }

      
      // call the batch URLs and populate the friends
      currentFriend = 0;
      for (int i = 0;i < blockCount;i++) {
      
//        String batchURL = "https://graph.facebook.com/";
//        HttpPost request = new HttpPost(batchURL);
//        String batchData = batchRequests[i];
//
//        // Create a new MultipartEntity for holding the POST data
//        MultipartEntity entity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE);
//        StringBody batchDataBody = new StringBody(batchData);
//        entity.addPart("batch",batchDataBody);
//        entity.addPart("access_token",new StringBody(FacebookAuthenticator.authClient.getAccessToken()));
//
//        request.setEntity(entity);
//
//        // Send the request
//        DefaultHttpClient client = new DefaultHttpClient();
//        HttpResponse response = client.execute(request);
//
//        // Get the response code
//        if (response.getStatusLine().getStatusCode() != HttpStatus.SC_OK) {
//          // Failure somewhere, so skip to the next batch
//          continue;
//        }
//
//        // Success
//        String responseMessage = EntityUtils.toString(response.getEntity());
//        JSONArray responses = new JSONArray(responseMessage);

        JSONArray responses = queryServerBatch(batchRequests[i]);

        for (int j = 0;j < 20;j++) {
          if (currentFriend >= friendCount) {
            break;
          }

          try {
            String batchObject = responses.getJSONObject(j).optString("body","");
            JSONObject checkinObject = new JSONObject(batchObject);

            JSONArray checkinsArray = checkinObject.getJSONArray("data");

            FacebookFriend friend = friends[currentFriend];

            if (checkinsArray.length() >= 1) {
              JSONObject checkinWrapper = checkinsArray.getJSONObject(0);

              // Location
              try {
                friend.setCheckin(checkinWrapper.getJSONObject("place").optString("name",""));
              }
              catch (JSONException e) {}

              // Timestamp (2011-07-02T00:45:38+0000)
              friend.setCheckinTime(checkinWrapper.optString("created_time",""));
            }
          }
          catch (JSONException e) {
            // Ignore any errors trying to set the check-in information
          }

          currentFriend++;
        }

      }

    }
    catch (Throwable t) {
      // Error in the response - maybe the server is down or the response message has changed
      String error = "error";
    }
  }*/
  
  
  /***********************************************************************************************
  Fetches the status for each friend
  @param friends the friends
  @param ids the user ids for each friend
  ***********************************************************************************************/
  /*public static void oldGetStatuses(FacebookFriend[] friends,long[] ids){
    try {

      if (friends == null || friends.length == 0 || ids == null || ids.length == 0) {
        return;
      }

      int friendCount = friends.length;
      if (friendCount != ids.length) {
        return;
      }

      // Create the batch requests (max of 20 per batch request)
      int blockCount = friendCount / 20;
      int extraCount = friendCount ^ 20;
      if (extraCount > 0) {
        blockCount++;
      }

      String[] batchRequests = new String[blockCount];
      int currentFriend = 0;
      for (int i = 0;i < blockCount;i++) {
        String batchData = "[";

        for (int j = 0;j < 20;j++) {
          if (currentFriend >= friendCount) {
            break;
          }

          if (j > 0) {
            batchData += ",";
          }

          batchData += "{ \"method\": \"GET\", \n" + 
                         "\"relative_url\": \"" + ids[currentFriend] + "/statuses?limit=1&fields=message,updated_time\", \n" + 
                        "}\n";
          currentFriend++;
        }
        batchData += "]";

        batchRequests[i] = batchData;
      }

      // call the batch URLs and populate the friends
      currentFriend = 0;
      for (int i = 0;i < blockCount;i++) {
//        String batchURL = "https://graph.facebook.com/";
//        HttpPost request = new HttpPost(batchURL);
//        String batchData = batchRequests[i];
//
//        // Create a new MultipartEntity for holding the POST data
//        MultipartEntity entity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE);
//        StringBody batchDataBody = new StringBody(batchData);
//        entity.addPart("batch",batchDataBody);
//        entity.addPart("access_token",new StringBody(FacebookAuthenticator.authClient.getAccessToken()));
//
//        request.setEntity(entity);
//
//        // Send the request
//        DefaultHttpClient client = new DefaultHttpClient();
//        HttpResponse response = client.execute(request);
//
//        // Get the response code
//        if (response.getStatusLine().getStatusCode() != HttpStatus.SC_OK) {
//          // Failure somewhere, so skip to the next batch
//          continue;
//        }
//
//        // Success
//        String responseMessage = EntityUtils.toString(response.getEntity());
//        JSONArray responses = new JSONArray(responseMessage);

        JSONArray responses = queryServerBatch(batchRequests[i]);

        for (int j = 0;j < 20;j++) {
          if (currentFriend >= friendCount) {
            break;
          }

          try {
            String batchObject = responses.getJSONObject(j).optString("body","");
            JSONObject statusObject = new JSONObject(batchObject);

            JSONArray statusArray = statusObject.getJSONArray("data");

            FacebookFriend friend = friends[currentFriend];

            if (statusArray.length() >= 1) {
              JSONObject statusWrapper = statusArray.getJSONObject(0);

              // Message
              friend.setStatusMessage(statusWrapper.optString("message",""));

              // Timestamp (2011-07-02T00:45:38+0000)
              friend.setStatusMessageTime(statusWrapper.optString("updated_time",""));
            }
          }
          catch (JSONException e) {
            // Ignore any errors trying to set the check-in information
          }

          currentFriend++;
        }

      }

    }
    catch (Throwable t) {
      // Error in the response - maybe the server is down or the response message has changed
      String error = "error";
    }
  }*/


  /***********************************************************************************************
  Signs a query, sends it to the server, receives a response and converts it to a <code>JSONObject</code> 
  @param queryUrl the URL of the query you wish to perform
  @return the response as a <code>JSONObject</code>, or <b>null</b> if there was an error
  ***********************************************************************************************/
  public static JSONObject queryServer(String queryUrl){
    try {

      // Check that we have valid authentication
      if (!FacebookAuthenticator.credentialsKnown()) {
        return null;
      }
      
      // Sign the request with the access token
      queryUrl = FacebookAuthenticator.signRequest(queryUrl);
      HttpGet request = new HttpGet(queryUrl);

      // Send the request
      DefaultHttpClient client = new DefaultHttpClient();
      HttpResponse response = client.execute(request);

      // Get the response code
      if (response.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
        // Success
        String responseMessage = EntityUtils.toString(response.getEntity());

        // return the JSONObject for the response
        return new JSONObject(responseMessage);
      }
      return null;
    }
    catch (Throwable t) {
      return null;
    }
  }

}
