////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.facebook.security;

import java.net.URLEncoder;


/***********************************************************************************************
Convenience methods for establishing OAuth 2.0 authentication to Facebook
***********************************************************************************************/
public class FacebookOAuth2Client extends OAuth2Client {

  /***********************************************************************************************
  Initialises the OAuth provider with the <i>consumerKey</i>, <i>consumerSecret</i> and the 
  <i>callbackUrl</i> for the application
  @param consumerKey the OAuth key for this application
  @param consumerSecret the OAuth secret for this application
  @param callbackUrl the OAuth callback URL for this application
  ***********************************************************************************************/
  public FacebookOAuth2Client(String consumerKey, String consumerSecret, String callbackUrl){
    super(consumerKey,consumerSecret,callbackUrl);
  }


  /***********************************************************************************************
  Gets the URL for converting an authentication <i>code</i> into an access token
  @param code the authentication code
  @return the URL used to generate the access token
  ***********************************************************************************************/
  public String getAccessTokenUrl(String code){
    setCode(code);
    String encodedCallbackUrl = URLEncoder.encode(callbackUrl);
    return "https://graph.facebook.com/oauth/access_token?client_id=" + consumerKey + "&redirect_uri=" + encodedCallbackUrl + "&client_secret=" + consumerSecret + "&code=" + code + "";
  }


  /***********************************************************************************************
  Gets the URL for authenticating the application
  @return the URL used to authenticate the application
  ***********************************************************************************************/
  public String getAuthenticationUrl(){
    String encodedCallbackUrl = URLEncoder.encode(callbackUrl);
    return "https://www.facebook.com/dialog/oauth?client_id=" + consumerKey + "&response_type=token&redirect_uri=" + encodedCallbackUrl + "&display=touch&scope=user_about_me,friends_about_me,friends_activities,friends_birthday,friends_checkins,friends_education_history,friends_events,friends_groups,friends_hometown,friends_interests,friends_likes,friends_location,friends_notes,friends_online_presence,friends_photo_video_tags,friends_photos,friends_relationships,friends_relationship_details,friends_religion_politics,friends_status,friends_videos,friends_website,friends_work_history,offline_access";
  }


  /***********************************************************************************************
  Converts the authentication <i>code</i> into an access token
  @param code the authentication code
  @return the access token
  ***********************************************************************************************/
  public String retreiveAccessToken(String code){
    
    if (code == null || code.equals("")){
      return "";
    }
    
    String token = "";
    
    // The code looks like this...
    // access_token=166942940015970%7C2.sa0&expires_in=64090
    // so need to extract the access token out of it.
    int tokenStart = code.indexOf("access_token=");
    if (tokenStart >= 0 && code.length() > 13){
      tokenStart += 13;
      int ampersandStart = code.indexOf("&",tokenStart);
      if (ampersandStart > tokenStart){
        token = code.substring(tokenStart,ampersandStart);
      }
    }
    
    setAccessToken(token);
    return token;

  }


  /***********************************************************************************************
  Signs a URL request with the access token (URLEncoded so that '|' in accessToken is converted to %7C)
  @param url the request to sign
  @return the signed request
  ***********************************************************************************************/
  public String signRequest(String url){
    if (url.indexOf('?') > 0) {
      // already has a '?', so need a '&' instead
      return url + "&oauth_token=" + URLEncoder.encode(getAccessToken());
    }
    else {
      return url + "?oauth_token=" + URLEncoder.encode(getAccessToken());
    }
  }

}
